import os

import pandas as pd

from nrsur_catalog import Catalog
from nrsur_catalog.utils import LATEX_LABELS
from .video_links import get_video_html
from .. import __website__

LINK = "<a href='{l}'> {txt}</a>"

ANIM_CELL = """
# + [markdown] tags=["full-width"]
# ## Animations  
# Here are some animations of the posterior.
# 
# |Spin       | 
# |-----------|
# | {spin}    |   
#
# |Remnant       | 
# |-----------|
# | {remnant}    | 
#
# -
"""


def get_animation_cell(event_name: str) -> str:
    """Returns the animation cell for the given event"""
    remnant_html = get_video_html(event_name, "remnant")
    spin_html = get_video_html(event_name, "spin")
    if remnant_html is None or spin_html is None:
        return ""
    return ANIM_CELL.format(spin=spin_html, remnant=remnant_html)


def get_catalog_summary(events_dir: str, cache_dir: str) -> pd.DataFrame:
    """
    Get a summary dataframe of the catalog
        events_dir: Dir with the notebooks
        cache_dir: Dir with the NRSurResults
    """
    catalog = Catalog.load(cache_dir)
    df = __load_processed_links_dataframe(catalog, events_dir)
    posterior_summary = __load_posterior_summary(catalog)

    # merge the posterior summary ('event') with the catalog summary ('event_name') columns
    df = df.merge(posterior_summary, on="event_id")
    df = df.sort_values(by="event_id")
    df = df.drop(columns=["event_id"])
    return df


def __check_fn(fname):
    if not os.path.isfile(fname):
        return ""
    return fname


def __thubnail(fname, event_link):
    if not os.path.isfile(fname):
        return "NA"
    thumbnail = f"{__website__}/_images/{os.path.basename(fname)}"
    thumbnail = f"<img src='{thumbnail}' width='100' height='50'>"
    thumbnail = LINK.format(l=event_link, txt=thumbnail)
    return thumbnail


def __load_processed_links_dataframe(catalog: Catalog, events_dir: str) -> pd.DataFrame:
    event_data = []
    for event in catalog.event_names:
        event_link = f"{__website__}/events/{event}.html"
        event_url = LINK.format(l=event_link, txt=event)
        event_data.append(
            {
                "event_id": event,
                "Event": event_url,
                "Waveform": __thubnail(
                    f"{events_dir}/{event}_waveform.png", event_link
                ),
            }
        )
    return pd.DataFrame(event_data)


def __load_posterior_summary(catalog):
    posterior_summary = catalog.get_latex_summary()
    posterior_summary["event_id"] = posterior_summary.index
    posterior_summary.index = range(len(posterior_summary))
    posterior_summary = posterior_summary.rename(
        columns={p: LATEX_LABELS.get(p, p) for p in posterior_summary.columns.values}
    )
    return posterior_summary


def is_file(f):
    return os.path.exists(f) or os.path.islink(f)



import pandas as pd

from nrsur_catalog.cache import CatalogCache
from nrsur_catalog import Catalog
import shutil, os

CLEANUP = False


def test_catalog_plot(mock_cache_dir, tmpdir):
    """Test that the Catalog class can be instantiated"""
    CACHE = CatalogCache(mock_cache_dir)

    # Test that the Catalog class can be instantiated from dir with results
    catalog = Catalog.load(CACHE.dir, clean=True)
    assert isinstance(catalog, Catalog)
    assert isinstance(catalog._df, pd.DataFrame)
    catalog.save(f"{tmpdir}/downsampled_catalog.h5")
    catalog = Catalog(posteriors=catalog._df)
    latex_summary = catalog.get_latex_summary()
    assert isinstance(latex_summary, pd.DataFrame)
    fig = catalog.violin_plot("mass_1")
    violin_fn = os.path.join(tmpdir, "violin_plot.png")
    fig.savefig(violin_fn)
    assert os.path.isfile(violin_fn)




"""Module to download, cache, and read the LVK posteriors"""
import h5py
import pandas as pd
from bilby.gw.result import CBCResult

from nrsur_catalog.api import download_event
from nrsur_catalog.cache import DEFAULT_CACHE_DIR, LVK_LABEL, CatalogCache

from .logger import logger


def load_lvk_result(event_name: str, cache_dir=DEFAULT_CACHE_DIR) -> CBCResult:
    """Load the LVK posterior for an event"""
    CACHE = CatalogCache(cache_dir)
    if not CACHE.find(event_name, lvk_posteriors=True):
        logger.debug(f"LVK {event_name} not in {CACHE.dir}. Downloading...")
        download_event(event_name, cache_dir=CACHE.dir, download_lvk=True)
    event_path = CACHE.find(event_name, lvk_posteriors=True, hard_fail=True)

    with h5py.File(event_path, "r") as f:
        res = f[LVK_LABEL]
        df = pd.DataFrame(res["posterior_samples"][:])

    return CBCResult(
        label="LVK [XPHM]",
        posterior=df,
        search_parameter_keys=list(df.columns),
    )



import unittest
from nrsur_catalog.api import zenodo_interface



def test_update_cache():
    zenodo_interface.cache_zenodo_urls_file(sandbox=True)

def test_names_of_events():
    events = zenodo_interface.get_zenodo_urls()
    event0 = list(events.keys())[0]
    present, name = zenodo_interface.check_if_event_in_zenodo(event0)
    assert present

def test_zenodo_check():
    assert zenodo_interface.check_if_event_in_zenodo("GW150914", lvk_posteriors=True)

if __name__ == "__main__":
    unittest.main()



import contextlib
import os
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bilby.gw.result import CompactBinaryCoalescenceResult
from bilby.gw.waveform_generator import WaveformGenerator

from .api import download_event
from .cache import DEFAULT_CACHE_DIR, CatalogCache
from .logger import logger
from .lvk_posterior import load_lvk_result
from .utils import (
    CATALOG_MAIN_COLOR,
    INTERESTING_PARAMETERS,
    LATEX_LABELS,
    get_1d_summary_str,
    get_dir_tree,
    pesummary_to_bilby_result,
    plot_overlaid_corner,
    prior_to_str,
)

pd.set_option("display.max_rows", None, "display.max_columns", None)


class NRsurResult(CompactBinaryCoalescenceResult):
    """Class to interface with a result of a NRSur Catalog analysis."""

    def __init__(self, *args, **kwargs):
        super(NRsurResult, self).__init__(*args, **kwargs)
        self._lvk_result = None

    @classmethod
    def load(
        cls,
        event_name: str,
        cache_dir: Optional[str] = DEFAULT_CACHE_DIR,
        event_path: Optional[str] = None,
    ) -> "NRsurResult":
        """Load a result (either by downloading or from a local dir).

        Parameters
        ----------
        event_name: str
            The name of the event to load
            (list all available events with `nrsur_catalog.catalog.Catalog.
        cache_dir: Optional[str]
            The credible level to use in the plot (default 0.9)
        event_path: Optional[str] = None
            The path to the NRSur result file to load.

        Returns
        -------
        NRsurResult: loaded result object.

        """

        CACHE = CatalogCache(cache_dir)
        if not CACHE.find(event_name) and event_path is None:
            logger.debug(f"{event_name} not in {CACHE.dir}. Downloading...")
            download_event(event_name, cache_dir)

        if event_path is None:
            event_path = CACHE.find(event_name)

        if event_path is None:
            raise ValueError(f"Event {event_name} not found.")

        r = pesummary_to_bilby_result(event_path)
        r.path_to_result = event_path
        r.outdir = os.path.join(CACHE.dir, event_name)
        r.label = event_name
        os.makedirs(r.outdir, exist_ok=True)
        r.__class__ = NRsurResult
        logger.debug(f"Loaded NRresult {event_name} from {event_path}")
        return r

    def summary(self, markdown: bool = False) -> Union[str, pd.DataFrame]:
        """Generate a table of the event parameter summary statistics.

        Parameters
        ----------
        markdown: bool
            True, return a string-markdown table of the summary
            False, return a pd.DataFrame of the summary.
        """

        selected_parameters = []
        summary_str = []
        priors = []

        for _, parameters in INTERESTING_PARAMETERS.items():
            for parameter in parameters:
                if parameter in self.posterior:
                    selected_parameters.append(LATEX_LABELS[parameter])
                    summary_str.append(get_1d_summary_str(self.posterior[parameter]))
                    pri = self.priors.get(parameter, "$-$")
                    if pri != "$-$":
                        pri = prior_to_str(pri)
                    priors.append(pri)

        df = pd.DataFrame(
            {
                "Parameter": selected_parameters,
                "Posterior 90% CI": summary_str,
                "Prior": priors,
            }
        )
        df.index = df["Parameter"]
        df = df.drop(columns=["Parameter"])
        if markdown:
            md_txt = df.to_markdown()
            md_txt = "\n".join([f"# {line}" for line in md_txt.splitlines()])
            return "\n" + md_txt
        else:
            return df

    def plot_skymap(self, *args, **kwargs):
        """Generate a skymap of the event"""
        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            # call the super class sky map method
            super(NRsurResult, self).plot_skymap()

    def __get_waveform_generator(self) -> WaveformGenerator:
        return self.waveform_generator_class(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            start_time=self.start_time,
            frequency_domain_source_model=self.frequency_domain_source_model,
            parameter_conversion=self.parameter_conversion,
            waveform_arguments=self.waveform_arguments,
        )

    def plot_signal(
        self,
        n_samples: Optional[int] = 1000,
        level: Optional[float] = 0.9,
        color: Optional[str] = CATALOG_MAIN_COLOR,
        polarisation: Optional[str] = "plus",
        outdir: Optional[str] = "",
    ) -> Union[plt.Figure, None]:
        """Generate a signal plot of the event

        Parameters
        ----------
        n_samples: int
            The number of samples to use in the plot
        level: float
            The credible level to use in the plot (default 0.9)
        color: str
            The color of the waveform.
        polarisation: str
            The polarisation to plot
        outdir:str
            The outdir to save the plot. If '' not saved.

        """

        if n_samples > len(self.posterior):
            n_samples = len(self.posterior)
            logger.warning(
                "n_samples > len(posterior), using n_samples = len(posterior)"
            )

        samples = self.posterior.sample(n_samples, replace=False)
        samples = samples.reset_index(drop=True)
        base_params = dict(samples.iloc[0])

        delta = (1 + level) / 2
        upper_percentile = delta * 100
        lower_percentile = (1 - delta) * 100
        ci = int(upper_percentile - lower_percentile)

        # Generate the plotting data
        waveform_generator = self.__get_waveform_generator()

        try:
            # suppress stdout
            with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
                base_wf = waveform_generator.time_domain_strain(base_params)[
                    polarisation
                ]
        except RuntimeError:
            logger.warning(
                "Unable to create a waveform: do you have NrSur7dq4 installed? Defaulting to IMRPhenomPv2"
            )
            self.waveform_arguments["waveform_approximant"] = "IMRPhenomPv2"
            self.waveform_arguments["minimum_frequency"] = 20
            waveform_generator = self.__get_waveform_generator()
            base_wf = waveform_generator.time_domain_strain(base_params)[polarisation]

        time_idx = np.arange(len(base_wf))
        waveforms = np.zeros((n_samples, len(base_wf)))

        for i, params in samples.iterrows():
            params = dict(params)
            waveforms[i] = waveform_generator.time_domain_strain(params)[polarisation]
        waveforms = np.roll(waveforms, 4 * len(base_wf) // 5, axis=1)

        median = np.median(waveforms, axis=0)
        lower, upper = np.quantile(
            waveforms, [lower_percentile / 100, upper_percentile / 100], axis=0
        )

        # Plot the data
        fig, ax = plt.subplots(1, 1, figsize=(6, 5))
        ax.plot(time_idx, median, color=color)
        ax.fill_between(
            time_idx,
            lower,
            upper,
            alpha=0.3,
            label=f"{ci}% credible region",
            color=color,
            linewidth=0,
        )
        ax.axis("off")
        outdir = self.outdir if outdir == "" else outdir
        filename = os.path.join(outdir, self.label + "_waveform.png")
        fig.tight_layout()
        fig.savefig(filename, dpi=300, bbox_inches="tight", transparent=True)
        return fig

    def plot_corner(
        self,
        parameters: Optional[List[str]] = None,
        priors: Optional[bool] = None,
        titles: Optional[List[str]] = True,
        save: Optional[bool] = False,
        filename: Optional[str] = None,
        dpi: Optional[int] = 300,
        **kwargs,
    ) -> Union[plt.Figure, None]:
        labels = []
        if parameters is not None:
            labels = [LATEX_LABELS[p] for p in parameters]
        if not "labels" in kwargs:
            kwargs["labels"] = labels
        if not "color" in kwargs:
            kwargs["color"] = CATALOG_MAIN_COLOR

        return super(NRsurResult, self).plot_corner(
            parameters, priors, titles, save, filename, dpi, **kwargs
        )

    def plot_lvk_comparison_corner(self, parameters: List[str], **kwargs):
        """Plot a corner plot comparing the posterior to the LVK catalog."""
        return plot_overlaid_corner(
            r1=self,
            r2=self.lvk_result,
            parameters=parameters,
            labels=["NRSur7dq4", "LVK [XPHM]"],
            colors=[CATALOG_MAIN_COLOR, "tab:blue"],
            **kwargs,
        )

    def print_configs(self) -> str:
        configs = self.meta_data["config_file"]
        for key, config in configs.items():
            k = f"{key}:"
            print(f"{k:<35} {config}")

    def download_analysis_datafiles(self, outdir: str):
        """Downloads all the analysis datafiles -- CURRENTLY NOT IMPLEMENTED.
        Please raise a git-issue if this is something you would like us to add.
        """
        if outdir is None:
            outdir = f"outdir_{self.label}"
        raise NotImplementedError(
            "Please raise a git-issue if this is something that you would "
            "really like us to include."
        )
        logger.info(f"Saving files to {outdir}")
        logger.info("To be implemented...")
        logger.info("Downloading analysis strain")
        logger.info("Downloading PSDs")
        logger.info("Downloading calibration files")
        logger.info("writing analysis config file")
        logger.info(f"Files:\n{get_dir_tree(outdir)}")

    @property
    def lvk_result(self) -> CompactBinaryCoalescenceResult:
        """The event's corresponding LVK result (from the GWTC)."""
        if not hasattr(self, "_lvk_result"):
            self._lvk_result = load_lvk_result(
                self.label, cache_dir=os.path.dirname(self.outdir)
            )
            logger.debug(f"Loaded LVKresult{self.label}")
        return self._lvk_result



"""Module to load all the catalog events and make plots with them"""
import os.path
from glob import glob
from itertools import chain
from typing import Dict, List, Optional, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from corner import hist2d
from matplotlib.ticker import AutoMinorLocator
from tqdm.auto import tqdm

from .api import download_missing_events, get_analysed_event_names
from .cache import DEFAULT_CACHE_DIR, NR_FILE_EXTENSION, CatalogCache
from .logger import logger
from .nrsur_result import NRsurResult
from .utils import (
    CATALOG_MAIN_COLOR,
    INTERESTING_PARAMETERS,
    LATEX_LABELS,
    format_qts_to_latex,
    get_event_name,
)

POSTERIORS_TO_KEEP = list(
    chain.from_iterable(v for v in INTERESTING_PARAMETERS.values())
)
# POSTERIORS_TO_KEEP += ["log_likelihood", "log_prior"]
MAX_SAMPLES = 10000
CATALOG_FN = "downsampled_posteriors.h5"

DEFAULT_CLEAN_CATALOG = True


class Catalog:
    def __init__(self, posteriors: pd.DataFrame):
        """Class to store the catalog events

        Parameters
        -----------
        posteriors: pd.DataFrame
            A dataframe with the catalog posteriors
            The dataframe must have an `event` column with the event name
        """
        self._df = posteriors
        # sort by event name (GW150914, GW151012, ...)
        self._df = self._df.sort_values("event")

    @classmethod
    def load(
        cls,
        cache_dir: Optional[str] = DEFAULT_CACHE_DIR,
        max_samples: Optional[int] = MAX_SAMPLES,
        clean: Optional[bool] = DEFAULT_CLEAN_CATALOG,
    ) -> "Catalog":
        """Load all the catalog posterior samples and stores a cache of the down-sampled posteriors.

        Parameters
        ----------
        cache_dir: Optional[str]
            The cache-dir where the NRSurrogate results will be loaded from.
            If empty, downloads all the NRSurrogate results.
        max_samples: Optional[int]
            The max number of samples to include from each event's posterior.
        clean: Optional[bool]
            Recomputes the cache of the downsampled-posteriors.
        """
        CACHE = CatalogCache(cache_dir)
        fname = f"{cache_dir}/{CATALOG_FN}"

        if not clean and os.path.isfile(fname):
            logger.info(f"Loading catalog from {fname}")
            catalog = cls(Catalog.from_hdf5(fname))
        else:
            if CACHE.is_incomplete:
                logger.warning(
                    f"Cache {cache_dir} is missing some events, downloading events..."
                )
                download_missing_events(CACHE.dir)
            df = Catalog.__load_events(cache_dir, max_samples=max_samples)
            catalog = cls(df)
            catalog.save(f"{cache_dir}/{CATALOG_FN}")
        return catalog

    @staticmethod
    def __load_events(events_dir: str, max_samples: int = MAX_SAMPLES) -> pd.DataFrame:
        """Helped method to load the posteriors from the Cache dir."""
        event_paths = glob(f"{events_dir}/*{NR_FILE_EXTENSION}")
        dfs = []
        assert len(event_paths) > 0, f"No events found in {events_dir}"
        event_paths = sorted(event_paths, key=get_event_name)
        for path in tqdm(event_paths, desc="Loading events"):
            event_name = get_event_name(path)

            r = NRsurResult.load(event_name, event_path=path)
            missing_params = set(POSTERIORS_TO_KEEP) - set(r.posterior.columns)
            if len(missing_params) > 0:
                logger.warning(
                    f"Event {event_name} does not have all the required parameters. Missing:"
                    f"{missing_params}. Adding mock data..."
                )
                for p in missing_params:
                    r.posterior[p] = np.nan
            posterior = r.posterior[POSTERIORS_TO_KEEP]
            if len(r.posterior) > max_samples:
                logger.debug(
                    f"Event {event_name} has {len(r.posterior)}>{max_samples} samples, "
                    f"downsampling to {max_samples} samples"
                )
                posterior = posterior.sample(
                    max_samples, weights=r.posterior.log_likelihood
                )
            posterior.loc[:, "event"] = event_name
            dfs.append(posterior)
        return pd.concat(dfs)

    def to_dict_of_posteriors(
        self, parameters: Optional[List[str]] = None
    ) -> Dict[str, pd.DataFrame]:
        """Return a dictionary of event:posteriors.

        Parameters
        ----------
        parameters: Optional[List[str]]
            List of the parameters to include in the posterior.


        Returns
        -------
        Dict[str:pd.DataFrame]: Returns a dictionary of {events: event_posteriors}.

        """
        if parameters is None:
            parameters = POSTERIORS_TO_KEEP
        return {
            e: self._df[self._df["event"] == e][parameters] for e in self.event_names
        }

    def save(self, path: str) -> None:
        """Cache the loaded catalog to a hdf5 file.

        Parameters
        ---------
        path: str
            Path to save the cache-catalog

        """
        assert path.endswith(".h5"), "Path must end with .h5"
        data = self.to_dict_of_posteriors()
        with h5py.File(path, "w") as f:
            for event_name, posterior in data.items():
                # write each posterior to a different key in the hdf5 file
                f.create_dataset(event_name, data=posterior.to_records(index=False))

    @staticmethod
    def from_hdf5(path: str) -> pd.DataFrame:
        """Load the catalog from a hdf5 file"""
        assert path.endswith(".h5"), "Path must end with .h5"
        with h5py.File(path, "r") as f:
            posteriors = []
            for key in f:
                df = pd.DataFrame(f[key][:])
                df["event"] = key
                posteriors.append(df)
            return pd.concat(posteriors)

    def violin_plot(self, parameter: str) -> plt.Figure:
        """Generate a violin plot of a parameter."""
        # posterior subsets for each event
        data = [
            post.values.ravel()
            for post in self.to_dict_of_posteriors([parameter]).values()
        ]
        means = [np.mean(d) for d in data]

        # sort by mean
        ylabels = [e for _, e in sorted(zip(means, self.event_names))]
        data = [d for _, d in sorted(zip(means, data), key=lambda x: x[0])]

        fig, ax = plt.subplots(figsize=(7, self.n * 0.8))
        ax.tick_params(axis="x", bottom=True, top=True, labelbottom=True, labeltop=True)
        for pos in ["right", "top", "bottom", "left"]:
            ax.spines[pos].set_visible(False)
        ax.set_xticks([])
        ax.set_title(LATEX_LABELS[parameter], fontsize=22)
        ax.set_xlabel(LATEX_LABELS[parameter], fontsize=22)
        ax.grid(True, axis="x", alpha=0.25)
        ax.xaxis.set_tick_params(width=0)
        ax.yaxis.set_tick_params(width=0)
        ax.set_yticks(range(1, self.n + 1))
        ax.set_yticklabels(ylabels)
        violin_parts = ax.violinplot(data, vert=False)
        for pc in violin_parts["bodies"]:
            pc.set_color(CATALOG_MAIN_COLOR)
        for partname in ("cbars", "cmins", "cmaxes"):
            vp = violin_parts[partname]
            vp.set_edgecolor(CATALOG_MAIN_COLOR)
        min_x = min([min(d) for d in data])
        max_x = max([max(d) for d in data])
        axes_ticks = np.linspace(min_x, max_x, 5)
        ax.set_ylim(0.5, self.n + 0.5)
        ax.set_xticks(axes_ticks)

        # annotate a group number every 7 y ticks
        ntics = 7
        for i in range(6, self.n, ntics):
            for xval in axes_ticks:
                ax.text(
                    xval,
                    i + 1.5,
                    f"{xval:.1f}",
                    fontsize=12,
                    ha="center",
                    va="center",
                    color="black",
                    alpha=0.25,
                )

        plt.tight_layout()
        fig.savefig(f"{parameter}_violin.png", dpi=300, bbox_inches="tight")
        return fig

    @property
    def parameters(self) -> List[str]:
        """Return a list of parameters in the catalog"""
        p = list(self._df.columns)
        p.remove("event")
        return p

    def get_posterior_quantiles(self, quantiles=[0.16, 0.5, 0.84]) -> pd.DataFrame:
        """Return posterior quantiles for each event's posteriors.

        Parameters
        ----------
        quantiles: List of floats marking the quantiles.

        """
        quant_df = self._df.groupby("event").quantile(quantiles)
        quant_df.index = quant_df.index.rename(["event", "quantile"])
        return quant_df

    @property
    def event_names(self) -> List[str]:
        """Return a list of loaded events in the catalog object"""
        return list(self._df.event.unique())

    @property
    def n(self) -> int:
        """Return the number of events in the catalog"""
        return len(self.event_names)

    def get_latex_summary(self) -> pd.DataFrame:
        """
        Return a dataframe with the median and 68% credible interval for each parameter
        """
        post_quantiles = self.get_posterior_quantiles()
        latex_summary = {}
        for event in self.event_names:
            event_data = {}
            for param in self.parameters:
                event_quants = post_quantiles.loc[event, param].values
                event_data[param] = format_qts_to_latex(*event_quants)
            latex_summary[event] = event_data

        latex_summary = pd.DataFrame(latex_summary).T
        latex_summary.index.name = "event"
        return latex_summary

    def plot_2d_posterior(
        self,
        parm_x: str,
        parm_y: str,
        scatter_medians: Optional[bool] = True,
        event_posteriors: Optional[bool] = False,
        event_quantiles: Optional[bool] = True,
        contour_all_posterior_samples: Optional[bool] = True,
        colors: Optional[Union[str, List[str]]] = CATALOG_MAIN_COLOR,
    ) -> plt.Figure:
        """
        Plot a 2D posterior for all events in the catalog

        Parameters
        ----------
        parm_x:str
            The parameter plotted along the x-axis (e.g. mass_1, chi_eff...)
        parm_y:str
            The parameter plotted along the y-axis (e.g. mass_1, chi_eff...)
        scatter_medians:Optional[bool]=True
            Scatter median values of each event.
        event_posteriors:Optional[bool]=False
            Outline the event posterior.
        event_quantiles:Optional[bool]=True
            Plot the 68% CI quantiles of each event's posterior.
        contour_all_posterior_samples:Optional[bool]=True
            Plot contours around all the posterior samples
            (regardless of which event each sample belongs to)
        colors:Optional[Union[str, List[str]]]
            The color to be used for each event.
            If a string, the same color will be used for each event.

        Returns
        -------
        plt.Figure: A matplotlib figure with the plot.
        """
        fig, ax = plt.subplots(figsize=(5, 5))
        if isinstance(colors, str):
            colors = [colors] * self.n

        if event_posteriors or contour_all_posterior_samples:
            if event_posteriors:
                linewidths = 0.2
            else:
                linewidths = 0

            for i, event in enumerate(self.event_names):
                data = self.get_event_posterior(event)[[parm_x, parm_y]]
                hist2d(
                    ax=ax,
                    x=data[parm_x].values,
                    y=data[parm_y].values,
                    color=colors[i],
                    bins=50,
                    zorder=-2,
                    alpha=0.1,
                    plot_datapoints=False,
                    plot_density=False,
                    plot_contours=True,
                    no_fill_contours=False,
                    fill_contours=True,
                    levels=[0, 1.0 - np.exp(-0.5 * 2.1 ** 2)],
                    smooth=1.1,
                    contourf_kwargs=dict(linewidths=linewidths),
                    contour_kwargs=dict(linewidths=linewidths),
                )

        if scatter_medians:
            posterior_quantiles = self.get_posterior_quantiles()
            posterior_quantiles = posterior_quantiles[[parm_x, parm_y]]
            for i, event in enumerate(self.event_names):
                x, y = posterior_quantiles.loc[event].values[1]
                xlow, ylow = posterior_quantiles.loc[event].values[0]
                xhigh, yhigh = posterior_quantiles.loc[event].values[2]
                ax.scatter(x, y, color=colors[i], s=0.5)
                if event_quantiles:
                    ax.errorbar(
                        x,
                        y,
                        xerr=[[x - xlow], [xhigh - x]],
                        yerr=[[y - ylow], [yhigh - y]],
                        fmt="none",
                        color=colors[i],
                        alpha=0.2,
                        capsize=0,
                    )

        # add minor ticks
        ax.set_xlim(min(self._df[parm_x]), max(self._df[parm_x]))
        ax.set_ylim(min(self._df[parm_y]), max(self._df[parm_y]))
        ax.tick_params(
            axis="both",
            which="both",
            direction="in",
            top=True,
            right=True,
            labelsize=12,
            pad=5,
        )
        ax.tick_params(axis="x", pad=10)
        # add minor ticks
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(axis="both", which="major", length=8)
        ax.tick_params(axis="both", which="minor", length=4)
        ax.set_xlabel(LATEX_LABELS[parm_x])
        ax.set_ylabel(LATEX_LABELS[parm_y])
        plt.tight_layout()
        return fig

    def get_all_posteriors(self) -> pd.DataFrame:
        """Return the dataframe of all the event posteriors concatenated together.
        The column 'event' denotes which 'event' the posterior sample belongs to.

        Returns
        -------
        pd.DataFrame: A dataframe of the all event's posterior samples.
        """
        return self._df.copy()

    def get_event_posterior(self, event_name: str) -> pd.DataFrame:
        """
        Return a dataframe with the posterior for a given event.

        Parameter
        ---------
        event_name:str
            The name of the event that will have its posterior returned.

        Returns
        -------
        pd.DataFrame: A dataframe of the event's posterior samples
        """
        return self._df[self._df["event"] == event_name]

    @staticmethod
    def get_analysed_event_names() -> List[str]:
        """Return the list of the analysed events available in the NRSurrogate Catalog.

        Returns
        -------
        List[str]: List of the analysed event names.
        """
        return get_analysed_event_names()

    @property
    def all_analysed_events_loaded(self) -> bool:
        """True if all analysed events have been loaded
        (this will be False when the Catalog object is created
        before all event results have been downloaded).
        """
        analysed_events = set(self.get_analysed_event_names())
        loaded_events = set(self.event_names)
        return len(analysed_events - loaded_events) == 0



VIDEO_IFRAME = """<iframe width="560" height="315" src="https://www.youtube.com/embed/{ID}?loop=1" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share" allowfullscreen></iframe>"""


REMNANT_VIDEO = dict(
    GW150914_095045="https://youtu.be/mkxggi9SM8M",
    DEFAULT="https://youtu.be/mkxggi9SM8M", # None
)

SPIN_VIDEO = dict(
    GW150914_095045="https://youtu.be/msRN00Spjbs",
    DEFAULT="https://youtu.be/msRN00Spjbs", # None
)


def __extract_id(l) -> str:
    """extract the id after last / from a link"""
    return l.split("/")[-1]


def get_video_html(event: str, type: str) -> str:
    """Returns the link for the given event"""
    if type == "remnant":
        link = REMNANT_VIDEO.get(event, REMNANT_VIDEO["DEFAULT"])
    elif type == "spin":
        link = SPIN_VIDEO.get(event, SPIN_VIDEO["DEFAULT"])
    else:
        raise ValueError(f"Unknown type: {type}")
    if link is None:
        return ""
    id = __extract_id(link)
    return VIDEO_IFRAME.format(ID=id)



"""Module to build individual pages for the website"""

import os
import shutil

import jupytext
import nbformat
from ploomber_engine import execute_notebook

from ..cache import DEFAULT_CACHE_DIR, CatalogCache
from ..logger import logger
from ..nrsur_result import NRsurResult
from .utils import is_file, get_animation_cell

HERE = os.path.dirname(__file__)
GW_PAGE_TEMPLATE = os.path.join(HERE, "page_templates/gw_notebook_template.py")
CATALOG_TEMPLATE = os.path.join(HERE, "page_templates/catalog_plots.py")
TABLE_PAGE_TEMPLATE = os.path.join(HERE, "page_templates/gw_menu_page.py")


def make_events_menu_page(outdir: str, cache: CatalogCache) -> None:
    """Writes the events menu page"""
    py_fname = f"{outdir}/events/gw_menu_page.py"
    os.makedirs(os.path.dirname(py_fname), exist_ok=True)
    events_dir = os.path.abspath(os.path.join(outdir, "events"))
    _replace_strings_from_file(
        TABLE_PAGE_TEMPLATE,
        {"{{EVENTS_DIR}}": events_dir, "{{CACHE_DIR}}": cache.dir},
        py_fname,
    )
    ipynb_fn = convert_py_to_ipynb(py_fname)
    return execute_notebook(
        ipynb_fn,
        ipynb_fn,
        cwd=outdir,
        progress_bar=False,
        verbose=False,
        save_profiling_data=False,
    )


def convert_py_to_ipynb(py_fn) -> str:
    """Converts a python file to a jupyter notebook"""
    ipynb_fn = py_fn.replace(".py", ".ipynb")
    template_py_pointer = jupytext.read(py_fn, fmt="py:light")
    jupytext.write(template_py_pointer, ipynb_fn)

    # ensure notebook is valid
    notebook = nbformat.read(ipynb_fn, as_version=4)
    nbformat.validate(notebook)
    os.remove(py_fn)
    return ipynb_fn


def make_gw_page(event_name: str, outdir: str, cache: CatalogCache):
    """Writes the GW event notebook and executes it"""
    md_fn = f"{outdir}/{event_name}.py"
    logger.debug(f"Making {event_name} page")
    nrsurr_res = NRsurResult.load(event_name, cache_dir=cache.dir)
    summary_md = nrsurr_res.summary(markdown=True)
    animation_md = get_animation_cell(event_name)
    _replace_strings_from_file(
        GW_PAGE_TEMPLATE,
        {
            "{{GW EVENT NAME}}": event_name,
            "{{SUMMARY_TABLE}}": summary_md,
            "{{ANIMATION_CELL}}": animation_md
        },
        md_fn,
    )
    ipynb_fn = convert_py_to_ipynb(md_fn)
    logger.debug(
        f"Executing GW{event_name}:\n"
        f"    - cwd: {outdir}\n"
        f"    - ipynb_fn: {ipynb_fn}\n"
        f"    - cache: {cache.dir}\n"
    )
    return execute_notebook(
        ipynb_fn,
        ipynb_fn,
        cwd=outdir,
        save_profiling_data=True,
        profile_memory=True,
        progress_bar=False,
        verbose=False,
    )


def _replace_strings_from_file(fname: str, replacements: dict, outfname: str) -> None:
    """Replaces strings in a file"""
    with open(fname, "r") as f:
        txt = f.read()
        for key, value in replacements.items():
            txt = txt.replace(key, value)
    with open(outfname, "w") as f:
        f.write(txt)


def make_catalog_page(outdir: str, cache: CatalogCache):
    """Writes the catalog notebook and executes it"""
    py_fname = f"{outdir}/catalog_plots.py"
    shutil.copyfile(CATALOG_TEMPLATE, py_fname)
    ipynb_fn = convert_py_to_ipynb(py_fname)

    tmp_cache = f"{outdir}/{DEFAULT_CACHE_DIR}"
    os.makedirs(tmp_cache, exist_ok=True)
    for fname in os.listdir(cache.dir):
        src = os.path.join(cache.dir, fname)
        dst = os.path.join(tmp_cache, fname)
        if not is_file(dst):
            os.symlink(src, dst)

    execute_notebook(
        ipynb_fn,
        f"{outdir}/catalog_plots.ipynb",
        cwd=outdir,
        save_profiling_data=True,
        profile_memory=True,
    )



from nrsur_catalog.utils import get_event_name


def test_get_event_name():
    assert get_event_name("GW170817_NRSur7dq4_merged_result.json") == "GW170817"
    assert get_event_name("GW170817_4.5_0.1.hdf5") == "GW170817"
    assert get_event_name("https://sandbox.zenodo.org/record/1164558/files/GW150914_result.json") == "GW150914"
    assert get_event_name(
        "https://zenodo.org/record/5546663/files/IGWN-GWTC3p0-v1-GW200209_085452_PEDataRelease_mixed_cosmo.h5"
    ) == "GW200209_085452"




import os
import re

import numpy as np
import requests
from bilby.core.prior import Prior
from tqdm.auto import tqdm

from ..logger import logger
from .constants import (
    CATALOG_MAIN_COLOR,
    INTERESTING_PARAMETERS,
    LATEX_LABELS,
    LOG_PARAMS,
)
from .overlaid_corner import plot_overlaid_corner
from .pesummary_result_to_bilby_result import pesummary_to_bilby_result


def get_size_of_file(filename: str) -> str:
    """Get the size of a file in human-readable format"""
    size = os.path.getsize(filename)
    for x in ["bytes", "KB", "MB", "GB", "TB"]:
        if size < 1024.0:
            return "%3.1f %s" % (size, x)
        size /= 1024.0


def format_qts_to_latex(q_lo: float, q_mid: float, q_hi: float) -> str:
    """Format quantiles to a LaTeX string"""
    q_m, q_p = q_mid - q_lo, q_hi - q_mid
    fmt = "{{0:{0}}}".format(".2f").format
    summary = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
    return summary.format(fmt(q_mid), fmt(q_m), fmt(q_p))


def get_1d_summary_str(x: np.ndarray, quantiles=[0.16, 0.5, 0.84]) -> str:
    q_lo, q_mid, q_hi = np.quantile(x, quantiles)
    return format_qts_to_latex(q_lo, q_mid, q_hi)


def get_event_name(s: str):
    """Get the event name from a string using a regex
    e.g.
    'GW170817' from 'GW170817_4.5_0.1.hdf5'
    'GW150914' from "https://sandbox.zenodo.org/record/1164558/files/GW150914_result.json"
    'GW200224_222234'  from  "GW200224_222234_NRSur7dq4_merged_result.json"
    """
    try:
        return re.findall(r"(GW\d{6}\_\d{6}|GW\d{6})", s)[0]
    except IndexError:
        logger.debug(f"Could not parse event name from {s}")
        return None


def get_dir_tree(path):
    """Get a directory tree (in a string)"""
    tree = ""
    for root, dirs, files in os.walk(path):
        level = root.replace(path, "").count(os.sep)
        indent = " " * 4 * (level)
        tree += f"{indent}{os.path.basename(root)}/\n"
        subindent = " " * 4 * (level + 1)
        for f in files:
            tree += f"{subindent}{f}\n"
    return tree


def download(url: str, fname: str) -> None:
    """Download a file from a URL and save it to a file"""
    resp = requests.get(url, timeout=5)
    if resp.status_code != 200:
        raise ValueError(f"Failed to download {url}: {resp.json()}")
    total = int(resp.headers.get("content-length", 0))
    with open(fname, "wb") as file, tqdm(
        desc=f"Downloading file",
        total=total,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            bar.update(size)


def prior_to_str(prior: Prior):
    """Convert a prior to a string"""
    # get class name
    name = prior.__class__.__name__
    min_val, max_val = None, None
    if hasattr(prior, "minimum"):
        min_val = round(prior.minimum, 2)
    if hasattr(prior, "maximum"):
        max_val = round(prior.maximum, 2)
    repr = r"$\text{" + name + "}"
    if min_val is not None and max_val is not None:
        repr += f" [{min_val}, {max_val}]"
    return repr + "$"


def safe_plot(func):
    """Decorator to catch errors when plotting"""

    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.error(f"Failed to plot {func.__name__}: {e}")

    return wrapper



import pytest
from nrsur_catalog.cache import CatalogCache


def test_cache(mock_cache_dir):
    CACHE = CatalogCache(mock_cache_dir)
    assert (len(CACHE.event_names) > 0)
    assert (mock_cache_dir == CACHE.dir)

    assert isinstance(CACHE.find(CACHE.event_names[0]), str)
    with pytest.raises(FileNotFoundError):
        CACHE.find(name="fake event", hard_fail=True)




"""Module to get zenodo URLs from NRSur Catlog zenodo page"""
import functools
import os
import re
from typing import Dict, List

from ..logger import logger
from ..utils import get_event_name

HERE = os.path.dirname(os.path.abspath(__file__))
LVK_URL_FILE = os.path.join(HERE, "lvk_urls.txt")
NR_URL_FILE = os.path.join(
    HERE, "nrsur_urls.txt"
)  # This file's contents are autogenerated
TITLE = "NRSurrogate Catalog Posteriors"


def _commit_url_file():
    try:
        import git

        repo_root = os.path.join(HERE, "../../..")
        repo = git.Repo(repo_root)
        repo.git.add(NR_URL_FILE)
        repo.git.commit(m=f"update urls [automated]")
        repo.git.push()
    except Exception as e:
        pass


def cache_zenodo_urls_file(sandbox=True) -> None:
    """Update the file URLs to the data on zenodo"""
    from zenodo_python import Deposition

    zeno = Deposition.from_title(title=TITLE, test=sandbox)
    logger.info(f"Zenodo {zeno} has {len(zeno.files)} files. Caching download URLs.")
    zeno.save_wget_file(NR_URL_FILE)
    file_contents = open(NR_URL_FILE, "r").read()
    logger.info(f"Finished caching Zenodo URLs to {NR_URL_FILE}:\n{file_contents}")
    _commit_url_file()


def get_zenodo_urls(lvk_posteriors=False) -> Dict[str, str]:
    """Returns a dictionary of the analysed events and their urls"""

    url_file = NR_URL_FILE
    if lvk_posteriors:
        url_file = LVK_URL_FILE

    # read in the zenodo_urls.txt file (each line is a url)
    with open(url_file, "r") as f:
        urls = f.readlines()
        urls = [url.strip() for url in urls]

    if len(urls) == 0:
        raise RuntimeError(f"No URLs found in {url_file}")

    # extract the event name from the url
    event_names = [get_event_name(url) for url in urls]
    return dict(zip(event_names, urls))


def check_if_event_in_zenodo(event_name: str, lvk_posteriors=False) -> bool:
    """Check if the event is in Zenodo"""

    present = False
    urls_dict = get_zenodo_urls(lvk_posteriors)

    shortname = event_name.split("_")[0]
    fullname = shortname + r"\_\d{6}"

    if event_name in urls_dict:
        present = True
    elif shortname in urls_dict:
        present = True
        event_name = shortname
    else:
        for name in urls_dict.keys():
            if re.match(fullname, name):
                present = True
                event_name = name
                break

    return present, event_name


@functools.lru_cache
def get_analysed_event_names(lvk_posteriors=False) -> List[str]:
    """Return a list of analysed events.

    Parameters
    ----------
        lvk_posteriors: Optional[bool]
            True if you want to return the names of the analysed LVK events.
            False if you want to return the names of the analysed NRSur cat events.
    """
    return list(set(get_zenodo_urls(lvk_posteriors).keys()))



import os.path
import pytest

from nrsur_catalog import NRsurResult
from nrsur_catalog.cache import CatalogCache


def test_load_result_from_cache(mock_cache_dir, mock_download):
    """Test that the GWResult class can be instantiated"""
    CACHE = CatalogCache(mock_cache_dir)
    event = CACHE.event_names[0]
    nrsur_result = NRsurResult.load(event, cache_dir=mock_cache_dir)
    nrsur_result.plot_corner(parameters=["mass_1", "mass_2"])
    nrsur_result.plot_lvk_comparison_corner(["luminosity_distance", "ra", "dec"])
    nrsur_result.plot_signal()

    # test that ValueError is raised if event is not in cache
    with pytest.raises(ValueError):
        event_name = "GW100000"
        nrsur_result.load(event_name)

    fpath = CACHE.find(CACHE.event_names[0])
    assert os.path.isfile(fpath)



from typing import List

from bilby.gw.result import CBCResult
from matplotlib.lines import Line2D

from .constants import LATEX_LABELS


def plot_overlaid_corner(
    r1: CBCResult,
    r2: CBCResult,
    parameters: List[str],
    labels: List[str],
    colors: List[str],
    **kwargs,
):
    # ensure parameters are in both results posteriors
    _check_list_is_subset(parameters, list(r1.posterior.columns))
    _check_list_is_subset(parameters, list(r2.posterior.columns))

    # plot corner
    latex = [LATEX_LABELS[p] for p in parameters]
    kwgs = dict(parameters=parameters, labels=latex, save=False, **kwargs)
    fig = r1.plot_corner(color=colors[0], titles=True, **kwgs)
    fig = r2.plot_corner(fig=fig, color=colors[1], quantiles=[], titles=False, **kwgs)

    # add legend to figure to right using the following colors
    legend_elements = [
        Line2D([0], [0], color=c, lw=4, label=l) for c, l in zip(colors, labels)
    ]
    fig.legend(
        handles=legend_elements,
        loc="upper right",
        bbox_to_anchor=(0.95, 0.95),
        bbox_transform=fig.transFigure,
        frameon=False,
        fontsize=16,
    )

    return fig


def _check_list_is_subset(l1, l2):
    s1 = set(l1)
    s2 = set(l2)
    if not s1.issubset(s2):
        raise ValueError(f"Missing parameters: {s1.difference(s2)}")



import unittest
from nrsur_catalog.web_builder import build_website
from nrsur_catalog.web_builder.make_pages import make_gw_page, make_events_menu_page
from nrsur_catalog.web_builder.utils import get_catalog_summary
import nrsur_catalog

nrsur_catalog.catalog.DEFAULT_CLEAN_CATALOG = False
from nrsur_catalog.cache import CatalogCache

import os
import glob
from unittest.mock import patch, PropertyMock

DIR = os.path.dirname(os.path.abspath(__file__))

CLEAN = True

def test_web(mock_cache_dir):
    cache = CatalogCache(mock_cache_dir)
    # generate
    outdir = "out_test_website"
    with patch("nrsur_catalog.cache.CatalogCache.is_incomplete", new_callable=PropertyMock) as mock_cache:
        mock_cache.return_value = False
        build_website(
            event_dir=mock_cache_dir,
            outdir=outdir,
            clean=CLEAN,
        )
    assert os.path.exists(outdir)
    html_dir = os.path.join(outdir, "_build/html")
    assert os.path.exists(f"{html_dir}/index.html")
    event_nbs = glob.glob(f"{html_dir}/_sources/events/GW*.ipynb")
    images = glob.glob(f"{html_dir}/_images/*waveform.png")
    assert len(event_nbs) == cache.num_events
    assert len(images) == cache.num_events
    print(f"ACCESS WEB HERE:\n{html_dir}/index.html")


def test_gw_page(mock_cache_dir, tmpdir):
    cache = CatalogCache(mock_cache_dir)
    name = cache.event_names[0]
    event_ipynb_dir = tmpdir
    make_gw_page(name, event_ipynb_dir, cache=cache)
    fpath = f"{event_ipynb_dir}/{name}.ipynb"
    assert os.path.exists(fpath)
    print(f"ACCESS GWPAGE HERE:\n {fpath}")


def test_menu_page(mock_cache_dir, tmpdir):
    cache = CatalogCache(mock_cache_dir)

    summary = get_catalog_summary(events_dir=tmpdir, cache_dir=cache.dir)
    assert len(summary) > 0

    event_ipynb_dir = tmpdir
    make_events_menu_page(outdir=event_ipynb_dir, cache=cache)
    assert os.path.exists(f"{event_ipynb_dir}/events/gw_menu_page.ipynb")


if __name__ == "__main__":
    unittest.main()



import os
import re
from glob import glob
from typing import List

from .api.zenodo_interface import get_analysed_event_names
from .logger import logger
from .utils import get_event_name

CACHE_ENV_VAR = "NRSUR_CATALOG_CACHE_DIR"
DEFAULT_CACHE_DIR = "./.nrsur_catalog_cache"
NR_FILE_EXTENSION = "_NRSur7dq4.h5"
LVK_FILE_EXTENSION = "_PEDataRelease_mixed_cosmo.h5"

NR_LABEL = "Bilby:NRSur7dq4"
LVK_LABEL = "C01:IMRPhenomXPHM"


class CatalogCache:
    """CatalogCache helps
    - check which files are in the cache
    - download any required new files
    """

    def __init__(self, cache_dir: str):
        """Class to handle the cache directory"""
        assert cache_dir is not None, "Cache directory must be specified"
        self._cache = cache_dir
        os.makedirs(self._cache, exist_ok=True)

    @property
    def dir(self) -> str:
        return os.path.abspath(self._cache)

    @property
    def list(self):
        return self._list()

    @property
    def list_lvk(self):
        return self._list(lvk_posteriors=True)

    def _list(self, lvk_posteriors=False) -> List[str]:
        """List the contents of the cache directory (sorted by number in filename)"""
        file_extension = LVK_FILE_EXTENSION if lvk_posteriors else NR_FILE_EXTENSION
        file_regex = os.path.join(self._cache, f"*{file_extension}")
        files = glob(file_regex)
        files = sorted(
            files, key=lambda x: int(re.findall(r"\d+", os.path.basename(x))[0])
        )
        return files

    @property
    def event_names(self) -> List[str]:
        """List the event names in the cache directory"""
        return [get_event_name(f) for f in self.list]

    @property
    def num_events(self) -> int:
        """Number of events in the cache directory"""
        return len(self.event_names)

    @property
    def event_names_lvk(self) -> List[str]:
        """List the event names in the cache directory"""
        return [get_event_name(f) for f in self.list_lvk]

    def find(self, name: str, hard_fail=False, lvk_posteriors=False) -> str:
        """Find a file in the cache directory"""
        file_extension = LVK_FILE_EXTENSION if lvk_posteriors else NR_FILE_EXTENSION
        fileregex = os.path.join(self._cache, f"*{name}*{file_extension}")
        filepath = glob(fileregex)
        if len(filepath) == 1:
            return filepath[0]
        if hard_fail:
            logger.debug(f"Current cache: {self.list} doesnt have {filepath}")
            raise FileNotFoundError(
                f"Could not find {name} in cache dir {self.dir}. "
                f"Is filepath:{filepath} present in the cache dir:"
                f"Listdir: {os.listdir(self.dir)}"
            )
        return ""

    def check_if_events_cached_in_zenodo(self, lvk_posteriors=False):
        """Return a list of events that are in the cache and in Zenodo"""
        zenodo_events = set(self.zenodo_events)
        local_events = set(self.event_names)
        if zenodo_events != local_events:
            logger.warning(
                f"Events in cache do not match events in Zenodo.\n"
                f"local: {local_events},\n"
                f"zenodo: {zenodo_events},\n"
                f"missing: {zenodo_events - local_events},\n"
                "Uploading the events to zenodo after building the website."
            )

    @property
    def is_empty(self) -> bool:
        return len(self.list) == 0

    @property
    def missing_events(self):
        zenodo_events = set(self.zenodo_events)
        local_events = set(self.event_names)
        return zenodo_events - local_events

    @property
    def is_incomplete(self) -> bool:
        """If not all events analysed are present"""
        return len(self.missing_events) != 0

    @property
    def zenodo_events(self) -> List[str]:
        if not hasattr(self, "__zenodo_events"):
            self.__zenodo_events = get_analysed_event_names()
        return self.__zenodo_events

    def __repr__(self):
        return f"CatalogCache(nrfiles={len(self.list)}, lvkfiles={len(self.list_lvk)}, cache_dir={self._cache})"



import logging
import os

import matplotlib.pyplot as plt

from .catalog import Catalog
from .nrsur_result import NRsurResult

__version__ = "0.0.2"
__website__ = "https://sxs-collaboration.github.io/"
__uri__ = "https://github.com/sxs-collaboration/nrsur_catalog"
__author__ = "NRSurrogate Catalog Team"
__email__ = "tousifislam24@gmail.com"
__license__ = "MIT"
__description__ = "GW event posteriors obtained using numerical relativity surrogate models"
__copyright__ = "Copyright 2023 NRSurrogate Catalog Team"
__contributors__ = "https://github.com/sxs-collaboration/nrsur_catalog/graphs/contributors"



HERE = os.path.dirname(__file__)

logging.getLogger("bilby").setLevel(logging.ERROR)

plt.style.use(f"{HERE}/style.mplstyle")



from nrsur_catalog.utils.pesummary_result_to_bilby_result import (
    _parse_prior,
    _parse_posterior,
    pesummary_to_bilby_result,
)

from bilby.gw.prior import BBHPriorDict
from bilby.gw.result import CBCResult
import pandas as pd


def test_prior_parser():
    p = {
        'a_1': ["Uniform(minimum=0, maximum=0.99, name='a_1', latex_label='$a_1$', unit=None, boundary=None)"],
        'a_2': ["Uniform(minimum=0, maximum=0.99, name='a_2', latex_label='$a_2$', unit=None, boundary=None)"],
        'azimuth': [
            "Uniform(minimum=0, maximum=6.283185307179586, name=None, latex_label='$\\\\epsilon$', unit=None, boundary='periodic')"],
        'chirp_mass': [
            "UniformInComponentsChirpMass(minimum=12, maximum=400, name='chirp_mass', latex_label='$\\\\mathcal{M}$', unit=None, boundary=None)"],
        'geocent_time': [
            "Uniform(minimum=1126259462.2910001, maximum=1126259462.491, name='geocent_time', latex_label='$t_c$', unit='$s$', boundary=None)"],
        'luminosity_distance': [
            'UniformSourceFrame(minimum=100.0, maximum=10000.0, cosmology=FlatLambdaCDM(H0=67.7 km / (Mpc s), Om0=0.307, Tcmb0=2.725 K, Neff=3.05, m_nu=[0.   0.   0.06] eV, Ob0=0.0486), name=\'luminosity_distance\', latex_label=\'$d_L$\', unit=Unit("Mpc"), boundary=None)'],
        'mass_ratio': [
            "UniformInComponentsMassRatio(minimum=0.167, maximum=1, name='mass_ratio', latex_label='$q$', unit=None, boundary=None)"],
        'phase': [
            "Uniform(minimum=0, maximum=6.283185307179586, name='phase', latex_label='$\\\\phi$', unit=None, boundary='periodic')"],
        'phi_12': [
            "Uniform(minimum=0, maximum=6.283185307179586, name='phi_12', latex_label='$\\\\Delta\\\\phi$', unit=None, boundary='periodic')"],
        'phi_jl': [
            "Uniform(minimum=0, maximum=6.283185307179586, name='phi_jl', latex_label='$\\\\phi_{JL}$', unit=None, boundary='periodic')"],
        'psi': [
            "Uniform(minimum=0, maximum=3.141592653589793, name='psi', latex_label='$\\\\psi$', unit=None, boundary='periodic')"],
        'theta_jn': [
            "Sine(minimum=0, maximum=3.141592653589793, name='theta_jn', latex_label='$\\\\theta_{JN}$', unit=None, boundary=None)"],
        'tilt_1': [
            "Sine(minimum=0, maximum=3.141592653589793, name='tilt_1', latex_label='$\\\\theta_1$', unit=None, boundary=None)"],
        'tilt_2': [
            "Sine(minimum=0, maximum=3.141592653589793, name='tilt_2', latex_label='$\\\\theta_2$', unit=None, boundary=None)"],
        'total_mass': ["Constraint(minimum=60, maximum=400, name='total_mass', latex_label='$M$', unit=None)"],
        'zenith': [
            "Sine(minimum=0, maximum=3.141592653589793, name=None, latex_label='$\\\\kappa$', unit=None, boundary=None)"]}
    pri = _parse_prior(p)
    assert isinstance(pri, BBHPriorDict)
    assert pri['chirp_mass'].minimum == 12
    assert pri['chirp_mass'].maximum == 400


def test_pesummary_to_bilby(mock_nrsur_result):
    r = pesummary_to_bilby_result(mock_nrsur_result)
    assert isinstance(r, CBCResult)
    assert isinstance(r.posterior, pd.DataFrame)
    assert isinstance(r.priors, BBHPriorDict)
    assert r.meta_data['likelihood']['time_marginalization'] == False


INTERESTING_PARAMETERS = {
    "Mass Parameters": ["mass_1", "mass_2", "chirp_mass", "mass_ratio"],
    "Spin Parameters": ["a_1", "a_2", "tilt_1", "tilt_2", "chi_eff", "chi_p"],
    "Localisation Parameters": [
        "ra",
        "dec",
        "geocent_time",
        "luminosity_distance",
    ],
    "Other Parameters": [
        "phase",
        "azimuth",
        "zenith",
        "psi",
        "phi_jl",
        "phi_12",
        "theta_jn",
    ],
    "Remnant Parameters": ["final_mass", "final_spin", "final_kick"],
}


CATALOG_MAIN_COLOR = "tab:orange"

LATEX_LABELS = dict(
    mass_1=r"$m_1\ [M_{\odot}]$",
    mass_2=r"$m_2\ [M_{\odot}]$",
    chirp_mass=r"$\mathcal{M}\ [M_{\odot}]$",
    mass_ratio=r"$q$",
    a_1=r"$a_1$",
    a_2=r"$a_2$",
    tilt_1=r"$\theta_1$",
    tilt_2=r"$\theta_2$",
    chi_eff=r"$\chi_{\mathrm{eff}}$",
    chi_p=r"$\chi_{\mathrm{p}}$",
    ra=r"$\alpha$",
    dec=r"$\delta$",
    geocent_time=r"$t_c\ [s]$",
    luminosity_distance=r"$d_L\ [Mpc]$",
    phase=r"$\phi$",
    azimuth=r"$\phi$",
    zenith=r"$\theta$",
    psi=r"$\psi$",
    phi_jl=r"$\phi_{\mathrm{JL}}$",
    phi_12=r"$\phi_{12}$",
    theta_jn=r"$\theta_{\mathrm{JN}}$",
    final_mass=r"$m_f\ [M_{\odot}]$",
    final_spin=r"$\chi_f$",
    final_kick=r"$v_{\rm kick}\ [{\rm km/s}]$",
)


LOG_PARAMS = ["mass_1", "mass_2"]



"""Module to build the website for the catalog
Most of the code has been grabbed from dfm/tess-atlas
"""

import argparse
import glob
import os
import shutil
import sys
from multiprocessing import cpu_count

from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map

from ..cache import DEFAULT_CACHE_DIR, CatalogCache
from ..logger import logger
from ..utils import get_event_name
from .make_pages import make_catalog_page, make_events_menu_page, make_gw_page
from .utils import is_file

HERE = os.path.dirname(__file__)
WEB_TEMPLATE = os.path.join(HERE, "website_template")


def build_website(
    event_dir: str, outdir: str, clean: bool = False, parallel_build=False
) -> None:
    """Build the website for the catalog"""
    logger.info(
        f"Building website [args: event_dir:{event_dir}, outdir:{outdir}, clean:{clean}]"
    )

    if clean:
        shutil.rmtree(outdir, ignore_errors=True)

    CACHE = CatalogCache(os.path.abspath(event_dir))

    # make symlink to the web cache directory
    web_cache = os.path.join(outdir, f"events/{DEFAULT_CACHE_DIR}/")

    # make each file in the cache directory a symlink to the web cache directory
    for file in os.listdir(CACHE.dir):
        src = os.path.join(CACHE.dir, file)
        dst = os.path.join(web_cache, file)
        dst_dir = os.path.dirname(dst)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir, exist_ok=True)
        if not is_file(dst):
            assert os.path.exists(src), f"File {src} (src) does not exist"
            os.symlink(src, dst)

    event_names = CACHE.event_names
    num_events = len(event_names)
    if num_events == 0:
        raise ValueError(f"No events found in the cache directory: {event_dir}")

    logger.info(f"Building website with {num_events} events: {event_names}")
    shutil.copytree(WEB_TEMPLATE, outdir, dirs_exist_ok=True)
    event_ipynb_dir = os.path.join(outdir, "events")
    regex = os.path.join(os.path.abspath(event_ipynb_dir), "GW*.ipynb")

    logger.debug(f"Checking {regex} for processed notebooks")
    events_processed = [get_event_name(f.split("/")[-1]) for f in glob.glob(regex)]
    events_not_processed = list(set(event_names) - set(events_processed))

    logger.debug(
        f"Events to make webpages for: {len(events_not_processed)}/{num_events}"
    )

    build_commands = [
        f"build_gwpage {name} {event_ipynb_dir} {CACHE.dir}"
        for name in events_not_processed
    ]
    if parallel_build:
        num_threads = cpu_count() // 2
        if num_events < num_threads:
            num_threads = num_events
        logger.info(f"Executing GW event notebooks with {num_threads} threads")
        process_map(
            os.system,
            build_commands,
            desc="Executing GW Notebooks",
            max_workers=num_threads,
            total=len(events_not_processed),
        )
    else:
        for cmd in tqdm(build_commands, desc="Executing GW notebooks"):
            os.system(cmd)

    logger.info("Executing events menu notebook")
    make_events_menu_page(outdir, CACHE)

    logger.info("Executing catalog notebook")
    make_catalog_page(outdir, CACHE)

    command = f"jupyter-book build {outdir}"
    os.system(command)


def gwpage_main():
    """Runs one GW notebook"""
    args = sys.argv[1:]
    name = args[0]
    event_ipynb_dir = args[1]
    data_dir = args[2]
    CACHE = CatalogCache(data_dir)
    logger.debug(
        f"Running GW notebook for {name} in {event_ipynb_dir} with data in {data_dir}"
    )
    make_gw_page(name, event_ipynb_dir, cache=CACHE)


def main():
    """Executes and builds the website [build_nrsur_website]"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--event-dir",
        type=str,
        default=".",
        help="Directory to load the events from",
    )
    parser.add_argument(
        "--outdir",
        type=str,
        default=".",
        help="Directory to write the website to",
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        help="Clean the output directory",
    )
    args = parser.parse_args()
    build_website(args.event_dir, args.outdir, args.clean)



import logging
import sys
import warnings

from loguru import logger

logger.remove(0)
logger.add(
    sys.stderr,
    format="|<blue>NURSUR CATALOG</blue>|{time:DD/MM HH:mm:ss}|{level}| <green>{message}</green> ",
    colorize=True,
    level="INFO",
)


warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=UserWarning)
logging.getLogger("matplotlib.font_manager").setLevel(logging.ERROR)



"""setup.py file for nrsur_catalog package."""
import codecs
import re
import os
import sys
from setuptools import find_packages, setup

NAME = "nrsur_catalog"
HERE = os.path.dirname(os.path.realpath(__file__))
META_PATH = os.path.join("src", NAME, "__init__.py")

# require python 3.8 or higher
if sys.version_info < (3, 8):
    raise RuntimeError("nrsur_catalog requires python 3.8 or higher")

def read(*parts):
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


def find_meta(meta, meta_file=read(META_PATH)):
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), meta_file, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


INSTALL_REQUIRES = [
    "matplotlib",
    "loguru",
    "bilby[gw]",
]
EXTRA_REQUIRE = dict(
    dev=[
        "jupyter-book>=0.13.2",
        # "zenodo_python @ git+https://github.com/avivajpeyi/zenodo-python.git@main#egg",
        "ploomber-engine>=0.0.23",
        "tabulate",
        "jupytext",  # for converting py-ipynb
        "ghp-import",  # for publishing to github pages
        "sphinx_inline_tabs",  # for tabs (https://sphinx-inline-tabs.readthedocs.io/en/latest/)
        "pytest",
        "GitPython",
        "sphinx-inline-tabs",
        "sphinxcontrib-bibtex",
        "plotly",
        "papermill",
        "nbconvert",
        'itables',
        'pytest-mock',
    ]
)

setup(
    name=NAME,
    version=find_meta("version"),
    author=find_meta("author"),
    author_email=find_meta("email"),
    maintainer=find_meta("author"),
    maintainer_email=find_meta("email"),
    url=find_meta("uri"),
    license=find_meta("license"),
    description=find_meta("description"),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={NAME: ["*/**.txt"]},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRA_REQUIRE,
    zip_safe=True,
    entry_points={
        "console_scripts": [
            f"get_nrsur_event={NAME}.api.download_event:main",
            f"build_nrsur_website={NAME}.web_builder.build_website:main",
            f"build_gwpage={NAME}.web_builder.build_website:gwpage_main",
        ]
    },
)



"""Module containing API to let users download NRSur Catlog events from Zenodo

NOTE:
    GWTC3: https://zenodo.org/record/5546663
    GWTC2.1: https://zenodo.org/record/6513631

"""
import argparse
import os.path
import sys
from typing import Optional, Tuple

from .. import utils
from ..cache import DEFAULT_CACHE_DIR, CatalogCache
from ..logger import logger
from .zenodo_interface import check_if_event_in_zenodo, get_zenodo_urls


def get_cli_args(args=None) -> Tuple[str, bool, str]:
    """Get the NRSur Catlog event name from the CLI and return it"""

    if args is None:
        args = sys.argv[1:]  # Get all args except the script name

    parser = argparse.ArgumentParser(prog="download_event")
    parser.add_argument(
        "event_name",
        type=str,
        default="",
        help="The name of the event to be downloaded (e.g. GW190521)",
    )
    parser.add_argument(
        "--cache-dir",
        type=str,
        default=DEFAULT_CACHE_DIR,
        help="The dir to cache the NRSur Catlog events",
    )
    parser.add_argument(
        "--all",
        action="store_true",
        default=False,
    )
    args = parser.parse_args(args=args)

    if args.all is False and args.event_name == "":
        raise ValueError("Either --all or --event-name must be specified")

    return args.event_name, args.all, args.cache_dir


def download_event(
    event_name: str,
    cache_dir: Optional[str] = DEFAULT_CACHE_DIR,
    download_lvk: bool = False,
) -> None:
    """Download the NRSur Catlog events from Zenodo given the event name"""

    CACHE = CatalogCache(cache_dir)

    present, event_name = check_if_event_in_zenodo(
        event_name, lvk_posteriors=download_lvk
    )

    if not present:
        logger.debug("Event not found in Zenodo")
        return

    if event_name in CACHE.find(
        event_name, lvk_posteriors=download_lvk
    ):  # Check if the event is already cached
        logger.debug(f"Fit {event_name} already downloaded")
        return

    url_dict = get_zenodo_urls(lvk_posteriors=download_lvk)
    if event_name not in url_dict:
        raise ValueError(
            f"{event_name} has not been analysed yet -- please choose from {list(url_dict.keys())}"
        )

    url = url_dict[event_name]
    fname = url.split("/")[-1]
    savepath = os.path.abspath(os.path.join(cache_dir, fname))
    catalog_name = "LVK" if download_lvk else "NRSur"
    logger.info(
        f"Downloading {event_name} from the {catalog_name} Catalog -> {savepath}..."
    )
    utils.download(url_dict[event_name], savepath)
    logger.info("Download completed!")


def download_missing_events(cache_dir: str) -> None:
    """Download all NRSur Catlog events from Zenodo that are not already present"""
    cache = CatalogCache(cache_dir)
    events_to_download = cache.missing_events
    logger.info(f"Downloading all {len(events_to_download)} events...")
    for event_name in events_to_download:
        download_event(event_name, cache_dir=cache_dir)



import re
from typing import Dict, List

import bilby
import h5py
import numpy as np
import pandas as pd
from bilby.core.utils.io import recursively_load_dict_contents_from_group
from bilby.gw.prior import BBHPriorDict
from bilby.gw.result import CBCResult


def pesummary_to_bilby_result(pesummary_result: str):
    """Convert a pesummary result to a bilby result

    parameters
    ----------
    pesummary_result: str
        The path to the pesummary result file

    returns a bilby result object
    """
    with h5py.File(pesummary_result, "r") as f:
        data = recursively_load_dict_contents_from_group(f, "/")

    data = data["Bilby:NRSur7dq4"]
    priors = _parse_prior(data["priors"]["analytic"])
    meta_data = _recursively_unrwap_dict(data["meta_data"]["other"])
    meta_data["likelihood"][
        "parameter_conversion"
    ] = bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
    meta_data["likelihood"][
        "frequency_domain_source_model"
    ] = bilby.gw.source.lal_binary_black_hole
    meta_data["likelihood"][
        "waveform_generator_class"
    ] = bilby.gw.waveform_generator.WaveformGenerator
    meta_data["likelihood"][
        "time_domain_source_model"
    ] = bilby.gw.source.lal_binary_black_hole

    return CBCResult(
        posterior=_parse_posterior(data["posterior_samples"]),
        priors=priors,
        meta_data=meta_data,
        search_parameter_keys=list(priors.keys()),
    )


def _parse_prior(p: Dict[str, List[str]]) -> BBHPriorDict:
    """Parse a prior string into a bilby prior object

    parameters
    ----------
    p: dict
        p = {
        'a_1': ["Uniform(minimum=0,...)"],
        'a_2': ["Uniform(minimum=0,...)"],
        ...
        }

    returns a bilby prior object
    """
    pri = {k: v[0] for k, v in p.items()}
    # convert 'UniformInComponentsChirpMass' to 'bilby.gw.prior.UniformInComponentsChirpMass'
    for k, v in pri.items():
        if "UniformInComponentsChirpMass" in v:
            full = "bilby.gw.prior.UniformInComponentsChirpMass"
            if full not in v:
                v = v.replace("UniformInComponentsChirpMass", full)
            pri[k] = v
        elif "UniformSourceFrame" in v:
            full = "bilby.gw.prior.UniformSourceFrame"
            v1 = v
            if full not in v:
                v1 = v.replace("UniformSourceFrame", full)
            regex = re.compile(r"cosmology=FlatLambdaCDM\((.*)\), name")
            v1 = regex.sub(r"name", v1)
            v1 = v1.replace("""Unit("Mpc")""", "'Mpc'")
            pri[k] = v1
        elif "UniformInComponentsMassRatio" in v:
            full = "bilby.gw.prior.UniformInComponentsMassRatio"
            if full not in v:
                v = v.replace(
                    "UniformInComponentsMassRatio",
                    full,
                )
            pri[k] = v
    pri = BBHPriorDict(pri)
    return pri


def _parse_posterior(post: np.ndarray) -> pd.DataFrame:
    return pd.DataFrame({name: post[name][:] for name in post.dtype.names})


def _recursively_unrwap_dict(d: Dict[str, List[str]]) -> Dict[str, str]:
    res = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = _recursively_unrwap_dict(v)

        if isinstance(v, np.ndarray) or isinstance(v, list):
            if len(v) == 1:
                v = v[0]

        if isinstance(v, str):
            if "None" in v:
                v = None
        res[k] = v

    return res



from nrsur_catalog.lvk_posterior import load_lvk_result
import pandas as pd
from nrsur_catalog import utils
from nrsur_catalog.logger import logger

def test_lvk_posterior_loader(mock_download, tmpdir):
    """Test that the LVK posterior can be loaded"""
    result = load_lvk_result("GW150914_095045", cache_dir=tmpdir)
    assert result is not None
    assert isinstance(result.posterior, pd.DataFrame)
    posterior = load_lvk_result("GW150914_095045", cache_dir=tmpdir) # get from cache
    assert posterior is not None


from glob import glob

import numpy as np
from bilby.gw.result import CBCResult
from tqdm import tqdm


def compute_ln_prior(result: CBCResult) -> np.ndarray:
    """Compute the log prior for a result object"""
    priors = result.priors
    priors_keys = list(priors.keys())
    # drop any keys with "recalib"
    priors_keys = [key for key in priors_keys if "recalib" not in key]
    samples = result.posterior[priors_keys]
    ln_prior = np.array([priors[key].ln_prob(samples[key]) for key in priors_keys])
    ln_prior = np.sum(ln_prior, axis=0)
    return ln_prior


def convert_result_json_to_hdf5(json_path: str):
    """Convert a result.json file to a result.hdf5 file.

    Parameters


    for key in result.posterior:
        print(key, result.posterior[key].dtype)

    ----------
    json_path: str
        The path to the result.json file.
    """
    result = CBCResult.from_json(json_path)
    result.sampling_time = result.sampling_time.total_seconds()
    result.posterior["log_prior"] = compute_ln_prior(result)
    col_to_drop = ["catch_waveform_errors", "waveform_approximant"]
    result.posterior = result.posterior.drop(columns=col_to_drop)
    for key in result.posterior:
        result.posterior[key] = np.array(result.posterior[key].tolist())
    hdf5_path = json_path.replace("result.json", "result.hdf5")
    result.save_to_file(hdf5_path, extension="hdf5")


def convert_all_result_json_to_hdf5(result_regex: str):
    """Convert all result.json files to result.hdf5 files"""
    for json_path in tqdm(glob(result_regex), desc="Converting"):
        convert_result_json_to_hdf5(json_path)
